# utility classes and functions
import numpy as np
from scipy.optimize import bisect


def projCappedSimplex(w, w_sum):
    a = np.min(w) - 1.0
    b = np.max(w) - 0.0

    def f(x):
        return np.sum(np.maximum(np.minimum(w - x, 1.0), 0.0)) - w_sum

    x = bisect(f, a, b)

    return np.maximum(np.minimum(w - x, 1.0), 0.0)



# nonlinear mixed effects model
from copy import deepcopy
from typing import Tuple

import cyipopt
import numpy as np
from numpy.typing import NDArray
from spmat.dlmat import BDLMat

from . import utils


class LimeTr:
    def __init__(
        self,
        n,
        k_beta,
        k_gamma,
        Y,
        F,
        JF,
        Z,
        S=None,
        share_obs_std=False,
        C=None,
        JC=None,
        c=None,
        H=None,
        JH=None,
        h=None,
        uprior=None,
        gprior=None,
        lprior=None,
        certain_inlier_id=None,
        inlier_percentage=1.0,
    ):
        """
        Create LimeTr object, for general mixed effects model

        Parameters
        ----------
        n : ndarray
            study sizes, n[i] is the number of observation for ith study.
        k_beta : int
            dimension of beta
        k_gamma : int
            dimension of gamma
        Y : ndarray
            study observations
        F : function
            return the predict observations given beta
        JF : function
            return the jacobian function of F
        Z : ndarray
            covariates matrix for the random effect
        S : optional, ndarray
            observation standard deviation
        """
        # pass in the dimension
        self.n = np.array(n)
        self.m = len(n)
        self.N = sum(n)
        self.k_beta = k_beta
        self.k_gamma = k_gamma
        # if include measurement error also as variable
        if S is not None:
            self.std_flag = 0
            self.k_delta = 0
        elif share_obs_std:
            self.std_flag = 1
            self.k_delta = 1
        else:
            self.std_flag = 2
            self.k_delta = self.m

        self.k = self.k_beta + self.k_gamma + self.k_delta
        self.k_total = self.k

        self.idx_beta = slice(0, self.k_beta)
        self.idx_gamma = slice(self.k_beta, self.k_beta + self.k_gamma)
        self.idx_delta = slice(self.k_beta + self.k_gamma, self.k)
        self.idx_split = np.cumsum(np.insert(n, 0, 0))[:-1]

        # pass in the data
        self.Y = Y
        self.F = F
        self.JF = JF
        self.Z = Z
        self.S = S
        if self.std_flag == 0:
            self.V = S**2

        # pass in the priors
        self.use_constraints = C is not None
        self.use_regularizer = H is not None
        self.use_uprior = uprior is not None
        self.use_gprior = gprior is not None
        self.use_lprior = lprior is not None

        self.C = C
        self.JC = JC
        self.c = c
        if self.use_constraints:
            self.constraints = C
            self.jacobian = JC
            self.num_constraints = C(np.zeros(self.k)).size
            self.cl = c[0]
            self.cu = c[1]
        else:
            self.num_constraints = 0
            self.cl = []
            self.cu = []

        self.H = H
        self.JH = JH
        self.h = h
        if self.use_regularizer:
            self.num_regularizer = H(np.zeros(self.k)).size
            self.hm = self.h[0]
            self.hw = 1.0 / self.h[1] ** 2
        else:
            self.num_regularizer = 0

        if self.use_uprior:
            self.uprior = uprior
        else:
            self.uprior = np.array(
                [
                    [-np.inf] * self.k_beta
                    + [0.0] * self.k_gamma
                    + [1e-7] * self.k_delta,
                    [np.inf] * self.k,
                ]
            )
            self.use_uprior = True

        self.lb = self.uprior[0]
        self.ub = self.uprior[1]

        if self.use_gprior:
            self.gprior = gprior
            self.gm = gprior[0]
            self.gw = 1.0 / gprior[1] ** 2

        if self.use_lprior:
            self.lprior = lprior
            self.lm = lprior[0]
            self.lw = np.sqrt(2.0) / lprior[1]

            # double dimension pass into ipopt
            self.k_total += self.k

            # extend the constraints matrix
            if self.use_constraints:

                def constraints(x):
                    v = x[: self.k]
                    v_abs = x[self.k :]

                    vec1 = C(v)
                    vec2 = np.hstack((v_abs - (v - self.lm), v_abs + (v - self.lm)))

                    return np.hstack((vec1, vec2))

                def jacobian(x):
                    v = x[: self.k]
                    v_abs = x[self.k :]
                    Id = np.eye(self.k)

                    mat1 = JC(v)
                    mat2 = np.block([[-Id, Id], [Id, Id]])

                    return np.vstack((mat1, mat2))

            else:

                def constraints(x):
                    v = x[: self.k]
                    v_abs = x[self.k :]

                    vec = np.hstack((v_abs - v, v_abs + v))

                    return vec

                def jacobian(x):
                    v = x[: self.k]
                    v_abs = x[self.k :]
                    Id = np.eye(self.k)

                    mat = np.block([[-Id, Id], [Id, Id]])

                    return mat

            self.num_constraints += 2 * self.k
            self.constraints = constraints
            self.jacobian = jacobian
            self.cl = np.hstack((self.cl, np.zeros(2 * self.k)))
            self.cu = np.hstack((self.cu, np.repeat(np.inf, 2 * self.k)))

            # extend the regularizer matrix
            if self.use_regularizer:

                def H_new(x):
                    v = x[: self.k]

                    return H(v)

                def JH_new(x):
                    v = x[: self.k]

                    return np.hstack((JH(v), np.zeros((self.num_regularizer, self.k))))

                self.H = H_new
                self.JH = JH_new

            # extend Uniform priors
            if self.use_uprior:
                uprior_abs = np.array([[0.0] * self.k, [np.inf] * self.k])
                self.uprior = np.hstack((self.uprior, uprior_abs))
                self.lb = self.uprior[0]
                self.ub = self.uprior[1]

        # trimming option
        self.use_trimming = 0.0 < inlier_percentage < 1.0
        self.certain_inlier_id = certain_inlier_id
        self.inlier_percentage = inlier_percentage
        self.num_inliers = np.floor(inlier_percentage * self.N)
        self.num_outliers = self.N - self.num_inliers
        self.w = np.repeat(self.num_inliers / self.N, self.N)

        if self.certain_inlier_id is not None:
            self.certain_inlier_id = np.unique(self.certain_inlier_id)
            self.active_trimming_id = np.array(
                [i for i in range(self.N) if i not in self.certain_inlier_id]
            )
        else:
            self.active_trimming_id = None

        # specify solution to be None
        self.soln = None
        self.info = None
        self.beta = np.zeros(self.k_beta)
        self.gamma = np.repeat(0.01, self.k_gamma)
        self.delta = np.repeat(0.01, self.k_delta)

        # check the input
        self.check()

    def check(self):
        assert self.Y.shape == (self.N,)
        assert self.Z.shape == (self.N, self.k_gamma)
        if self.S is not None:
            assert self.S.shape == (self.N,)
            assert np.all(self.S > 0.0)

        if self.use_constraints:
            assert self.c.shape == (2, self.num_constraints)
            assert np.all(self.cl <= self.cu)

        if self.use_regularizer:
            assert self.h.shape == (2, self.num_regularizer)
            assert np.all(self.h[1] > 0.0)

        if self.use_uprior:
            assert np.all(self.lb <= self.ub)

        if self.use_gprior:
            assert np.all(self.gprior[1] > 0.0)

        assert 0.0 < self.inlier_percentage <= 1.0
        if self.use_trimming and self.certain_inlier_id is not None:
            assert isinstance(self.certain_inlier_id, np.ndarray)
            assert self.certain_inlier_id.dtype == int
            assert self.certain_inlier_id.ndim == 1
            assert self.certain_inlier_id.size <= self.num_inliers
            assert np.min(self.certain_inlier_id) >= 0
            assert np.max(self.certain_inlier_id) < self.N

        if self.k > self.N:
            print("Warning: information insufficient!")

    def _get_vars(self, x: NDArray) -> Tuple[NDArray, NDArray, NDArray]:
        beta, gamma, delta = x[self.idx_beta], x[self.idx_gamma], x[self.idx_delta]
        gamma = np.maximum(0.0, gamma)
        delta = np.maximum(0.0, delta)
        return beta, gamma, delta

    def objective(self, x):
        # unpack variable
        beta, gamma, delta = self._get_vars(x)

        # trimming option
        if self.use_trimming:
            sqrt_w = np.sqrt(self.w)
            sqrt_W = sqrt_w.reshape(self.N, 1)
            F_beta = self.F(beta) * sqrt_w
            Y = self.Y * sqrt_w
            Z = self.Z * sqrt_W
            if self.std_flag == 0:
                V = self.V**self.w
            elif self.std_flag == 1:
                V = np.repeat(delta[0], self.N) ** self.w
            elif self.std_flag == 2:
                V = np.repeat(delta, self.n) ** self.w
        else:
            F_beta = self.F(beta)
            Y = self.Y
            Z = self.Z
            if self.std_flag == 0:
                V = self.V
            elif self.std_flag == 1:
                V = np.repeat(delta[0], self.N)
            elif self.std_flag == 2:
                V = np.repeat(delta, self.n)

        # residual and variance
        R = Y - F_beta
        D = BDLMat(dvecs=V, lmats=Z * np.sqrt(gamma), dsizes=self.n)

        val = 0.5 * self.N * np.log(2.0 * np.pi)
        val += 0.5 * D.logdet()
        val += 0.5 * R.dot(D.invdot(R))

        # add gpriors
        if self.use_regularizer:
            val += 0.5 * self.hw.dot((self.H(x) - self.hm) ** 2)

        if self.use_gprior:
            val += 0.5 * self.gw.dot((x[: self.k] - self.gm) ** 2)

        if self.use_lprior:
            val += self.lw.dot(x[self.k :])

        return val

    def gradient(self, x):
        # unpack variable
        beta, gamma, delta = self._get_vars(x)

        # trimming option
        if self.use_trimming:
            sqrt_w = np.sqrt(self.w)
            sqrt_W = sqrt_w.reshape(self.N, 1)
            F_beta = self.F(beta) * sqrt_w
            JF_beta = self.JF(beta) * sqrt_W
            Y = self.Y * sqrt_w
            Z = self.Z * sqrt_W
            if self.std_flag == 0:
                V = self.V**self.w
            elif self.std_flag == 1:
                V = np.repeat(delta[0], self.N) ** self.w
            elif self.std_flag == 2:
                V = np.repeat(delta, self.n) ** self.w
        else:
            F_beta = self.F(beta)
            JF_beta = self.JF(beta)
            Y = self.Y
            Z = self.Z
            if self.std_flag == 0:
                V = self.V
            elif self.std_flag == 1:
                V = np.repeat(delta[0], self.N)
            elif self.std_flag == 2:
                V = np.repeat(delta, self.n)

        # residual and variance
        R = Y - F_beta
        D = BDLMat(dvecs=V, lmats=Z * np.sqrt(gamma), dsizes=self.n)

        # gradient for beta
        DR = D.invdot(R)
        g_beta = -JF_beta.T.dot(DR)

        # gradient for gamma
        DZ = D.invdot(Z)
        g_gamma = 0.5 * np.sum(Z * DZ, axis=0) - 0.5 * np.sum(
            np.add.reduceat(DZ.T * R, self.idx_split, axis=1) ** 2, axis=1
        )

        # gradient for delta
        if self.std_flag == 0:
            g_delta = np.array([])
        elif self.std_flag == 1:
            d = -(DR**2) + D.invdiag()
            if self.use_trimming:
                v = np.repeat(delta[0], self.N)
                d *= self.w * (v ** (self.w - 1.0))
            g_delta = 0.5 * np.array([np.sum(d)])
        elif self.std_flag == 2:
            d = -(DR**2) + D.invdiag()
            if self.use_trimming:
                v = np.repeat(delta, self.n)
                d *= self.w * (v ** (self.w - 1.0))
            g_delta = 0.5 * (np.add.reduceat(d, self.idx_split))

        g = np.hstack((g_beta, g_gamma, g_delta))

        # add gradient from the regularizer
        if self.use_regularizer:
            g += self.JH(x).T.dot((self.H(x) - self.hm) * self.hw)

        # add gradient from the gprior
        if self.use_gprior:
            g += (x[: self.k] - self.gm) * self.gw

        # add gradient from the lprior
        if self.use_lprior:
            g = np.hstack((g, self.lw))

        return g

    def objectiveTrimming(self, w):
        t = (self.Z**2).dot(self.gamma)
        r = self.Y - self.F(self.beta)
        if self.std_flag == 0:
            v = self.V
        elif self.std_flag == 1:
            v = np.repeat(self.delta[0], self.N)
        elif self.std_flag == 2:
            v = np.repeat(self.delta, self.n)
        d = v + t

        val = 0.5 * np.sum(r**2 * w / d)
        val += 0.5 * self.N * np.log(2.0 * np.pi) + 0.5 * w.dot(np.log(d))

        return val

    def gradientTrimming(self, w, use_ad=False, eps=1e-10):
        if use_ad:
            # only use when testing
            g = np.zeros(self.N)
            z = w + 0j
            for i in range(self.N):
                z[i] += eps * 1j
                g[i] = self.objectiveTrimming(z).imag / eps
                z[i] -= eps * 1j

            return g

        t = (self.Z**2).dot(self.gamma)
        r = (self.Y - self.F(self.beta)) ** 2
        if self.std_flag == 0:
            v = self.V
        elif self.std_flag == 1:
            v = np.repeat(self.delta[0], self.N)
        elif self.std_flag == 2:
            v = np.repeat(self.delta, self.n)
        d = v + t

        g = 0.5 * r / d
        g += 0.5 * np.log(d)

        return g

    def optimize(
        self,
        x0=None,
        print_level=0,
        max_iter=100,
        tol=1e-8,
        acceptable_tol=1e-6,
        nlp_scaling_method=None,
        nlp_scaling_min_value=None,
    ):
        if x0 is None:
            x0 = np.hstack((self.beta, self.gamma, self.delta))
            if self.use_lprior:
                x0 = np.hstack((x0, np.zeros(self.k)))

        assert x0.size == self.k_total

        opt_problem = cyipopt.Problem(
            n=int(self.k_total),
            m=int(self.num_constraints),
            problem_obj=self,
            lb=self.uprior[0],
            ub=self.uprior[1],
            cl=self.cl,
            cu=self.cu,
        )

        opt_problem.add_option("print_level", print_level)
        opt_problem.add_option("max_iter", max_iter)
        opt_problem.add_option("tol", tol)
        opt_problem.add_option("acceptable_tol", acceptable_tol)
        if nlp_scaling_method is not None:
            opt_problem.add_option("nlp_scaling_method", nlp_scaling_method)
        if nlp_scaling_min_value is not None:
            opt_problem.add_option("nlp_scaling_min_value", nlp_scaling_min_value)

        soln, info = opt_problem.solve(x0)

        self.soln = soln
        self.info = info
        self.beta = soln[self.idx_beta]
        self.gamma = soln[self.idx_gamma]
        self.delta = soln[self.idx_delta]

    def fitModel(
        self,
        x0=None,
        inner_print_level=0,
        inner_max_iter=20,
        inner_tol=1e-8,
        inner_acceptable_tol=1e-6,
        inner_nlp_scaling_method=None,
        inner_nlp_scaling_min_value=None,
        outer_verbose=False,
        outer_max_iter=100,
        outer_step_size=1.0,
        outer_tol=1e-6,
        normalize_trimming_grad=False,
    ):
        if not self.use_trimming:
            self.optimize(
                x0=x0,
                print_level=inner_print_level,
                max_iter=inner_max_iter,
                acceptable_tol=inner_acceptable_tol,
                nlp_scaling_method=inner_nlp_scaling_method,
                nlp_scaling_min_value=inner_nlp_scaling_min_value,
            )

            return self.beta, self.gamma, self.w

        self.soln = x0

        num_iter = 0
        err = outer_tol + 1.0

        while err >= outer_tol:
            self.optimize(
                x0=self.soln,
                print_level=inner_print_level,
                max_iter=inner_max_iter,
                tol=inner_tol,
                acceptable_tol=inner_acceptable_tol,
                nlp_scaling_method=inner_nlp_scaling_method,
                nlp_scaling_min_value=inner_nlp_scaling_min_value,
            )

            w_grad = self.gradientTrimming(self.w)
            if normalize_trimming_grad:
                w_grad /= np.linalg.norm(w_grad)
            w_new = utils.projCappedSimplex(
                self.w - outer_step_size * w_grad,
                self.num_inliers,
                active_id=self.active_trimming_id,
            )

            err = np.linalg.norm(w_new - self.w) / outer_step_size
            np.copyto(self.w, w_new)

            num_iter += 1

            if outer_verbose:
                obj = self.objectiveTrimming(self.w)
                print("iter %4d, obj %8.2e, err %8.2e" % (num_iter, obj, err))

            if num_iter >= outer_max_iter:
                print("reach max outer iter")
                break

        return self.beta, self.gamma, self.w

    def estimateRE(self):
        """
        estimate random effect after fitModel
        """
        if self.soln is None:
            print("Please fit the model first.")
            return None

        if self.std_flag == 0:
            S = self.S
        elif self.std_flag == 1:
            S = np.sqrt(np.repeat(self.delta[0], self.N))
        elif self.std_flag == 2:
            S = np.sqrt(np.repeat(self.delta, self.n))

        if self.use_trimming:
            R = (self.Y - self.F(self.beta)) * np.sqrt(self.w)
            Z = self.Z * (np.sqrt(self.w).reshape(self.N, 1))
        else:
            R = self.Y - self.F(self.beta)
            Z = self.Z

        iV = 1.0 / S**2
        iVZ = Z * iV.reshape(iV.size, 1)

        r = np.split(R, np.cumsum(self.n)[:-1])
        v = np.split(S**2, np.cumsum(self.n)[:-1])
        z = np.split(Z, np.cumsum(self.n)[:-1], axis=0)
        ivz = np.split(iVZ, np.cumsum(self.n)[:-1], axis=0)

        u = []
        for i in range(self.m):
            rhs = ivz[i].T.dot(r[i])
            tmp = z[i] * self.gamma
            mat = np.diag(v[i]) + tmp.dot(z[i].T)
            vec = self.gamma * rhs - tmp.T.dot(np.linalg.solve(mat, tmp.dot(rhs)))
            u.append(vec)

        self.u = np.vstack(u)

        return self.u

    def estimate_re(
        self, beta: np.ndarray = None, gamma: np.ndarray = None, use_gamma: bool = True
    ) -> np.ndarray:
        beta = self.beta if beta is None else beta
        gamma = self.gamma if gamma is None else gamma
        r = np.split(self.Y - self.F(beta), np.cumsum(self.n)[:-1])
        z = np.split(self.Z, np.cumsum(self.n)[:-1], axis=0)
        v = np.split(self.S**2, np.cumsum(self.n)[:-1])

        u = []
        for i in range(self.m):
            rhs = (z[i].T / v[i]).dot(r[i])
            if use_gamma:
                q = (z[i].T / v[i]).dot(z[i]) * gamma + np.identity(self.k_gamma)
                u.append(gamma[:, None] * np.linalg.inv(q).dot(rhs))
            else:
                q = (z[i].T / v[i]).dot(z[i])
                u.append(np.linalg.inv(q).dot(rhs))

        return np.vstack(u)

    def get_gamma_fisher(self, gamma: np.ndarray) -> np.ndarray:
        z = np.split(self.Z, np.cumsum(self.n)[:-1], axis=0)
        v = np.split(self.S**2, np.cumsum(self.n)[:-1])
        H = np.zeros((self.k_gamma, self.k_gamma))
        for i in range(self.m):
            q = np.diag(v[i]) + (z[i] * gamma).dot(z[i].T)
            q = z[i].T.dot(np.linalg.inv(q).dot(z[i]))
            H += 0.5 * (q**2)
        return H

    def simulateData(self, beta_t, gamma_t, sim_prior=True, sim_re=True):
        # sample random effects and measurement error
        if sim_re:
            u = np.random.randn(self.m, self.k_gamma) * np.sqrt(gamma_t)
        else:
            if not hasattr(self, "u"):
                self.estimateRE()
            u = self.u

        U = np.repeat(u, self.n, axis=0)
        ZU = np.sum(self.Z * U, axis=1)

        if self.std_flag == 0:
            S = self.S
        elif self.std_flag == 1:
            S = np.sqrt(np.repeat(self.delta[0], self.N))
        elif self.std_flag == 2:
            S = np.sqrt(np.repeat(self.delta, self.n))

        E = np.random.randn(self.N) * S

        self.Y = self.F(beta_t) + ZU + E

        if sim_prior:
            if self.use_gprior:
                valid_id = ~np.isinf(self.gprior[1])
                valid_num = np.sum(valid_id)
                self.gm = np.zeros(self.k)
                self.gm[valid_id] = (
                    self.gprior[0][valid_id]
                    + np.random.randn(valid_num) * self.gprior[1][valid_id]
                )

            if self.use_regularizer:
                valid_id = ~np.isinf(self.h[1])
                valid_num = np.sum(valid_id)
                self.hm = np.zeros(self.num_regularizer)
                self.hm[valid_id] = (
                    self.h[0][valid_id]
                    + np.random.randn(valid_num) * self.h[1][valid_id]
                )

    @classmethod
    def testProblem(
        cls,
        use_trimming=False,
        use_constraints=False,
        use_regularizer=False,
        use_uprior=False,
        use_gprior=False,
        know_obs_std=True,
        share_obs_std=False,
    ):
        m = 10
        n = [5] * m
        N = sum(n)
        k_beta = 3
        k_gamma = 2
        if know_obs_std:
            k_delta = 0
        elif share_obs_std:
            k_delta = 1
        else:
            k_delta = m
        k = k_beta + k_gamma + k_delta

        beta_t = np.random.randn(k_beta)
        gamma_t = np.random.rand(k_gamma) * 0.09 + 0.01

        X = np.random.randn(N, k_beta)
        Z = np.random.randn(N, k_gamma)

        S = np.random.rand(N) * 0.09 + 0.01
        V = S**2
        D = np.diag(V) + (Z * gamma_t).dot(Z.T)

        U = np.random.multivariate_normal(np.zeros(N), D)
        E = np.random.randn(N) * S

        Y = X.dot(beta_t) + U + E

        def F(beta, X=X):
            return X.dot(beta)

        def JF(beta, X=X):
            return X

        # constraints, regularizer and priors
        if use_constraints:
            M = np.ones((1, k))

            def C(x, M=M):
                return M.dot(x)

            def JC(x, M=M):
                return M

            c = np.array([[0.0], [1.0]])
        else:
            C, JC, c = None, None, None

        if use_regularizer:
            M = np.ones((1, k))

            def H(x, M=M):
                return M.dot(x)

            def JH(x, M=M):
                return M

            h = np.array([[0.0], [2.0]])
        else:
            H, JH, h = None, None, None

        if use_uprior:
            uprior = np.array([[0.0] * k, [np.inf] * k])
        else:
            uprior = None

        if use_gprior:
            gprior = np.array([[0.0] * k, [2.0] * k])
        else:
            gprior = None

        if use_trimming:
            inlier_percentage = 0.9
        else:
            inlier_percentage = 1.0

        if not know_obs_std:
            S = None

        return cls(
            n,
            k_beta,
            k_gamma,
            Y,
            F,
            JF,
            Z,
            S=S,
            C=C,
            JC=JC,
            c=c,
            H=H,
            JH=JH,
            h=h,
            uprior=uprior,
            gprior=gprior,
            inlier_percentage=inlier_percentage,
            share_obs_std=share_obs_std,
        )

    @classmethod
    def testProblemLasso(cls):
        m = 100
        n = [1] * m
        N = sum(n)
        k_beta = 150
        k_gamma = 1
        k = k_beta + k_gamma

        beta_t = np.zeros(k_beta)
        beta_t[np.random.choice(k_beta, 5)] = np.sign(np.random.randn(5))
        gamma_t = np.zeros(k_gamma)

        X = np.random.randn(N, k_beta)
        Z = np.ones((N, k_gamma))
        Y = X.dot(beta_t)
        S = np.repeat(1.0, N)

        weight = 0.1 * np.linalg.norm(X.T.dot(Y), np.inf)

        def F(beta):
            return X.dot(beta)

        def JF(beta):
            return X

        uprior = np.array([[-np.inf] * k_beta + [0.0], [np.inf] * k_beta + [0.0]])
        lprior = np.array([[0.0] * k, [np.sqrt(2.0) / weight] * k])

        return cls(n, k_beta, k_gamma, Y, F, JF, Z, S=S, uprior=uprior, lprior=lprior)

    @staticmethod
    def sampleSoln(
        lt, sample_size=1, print_level=0, max_iter=100, sim_prior=True, sim_re=True
    ):
        beta_samples = np.zeros((sample_size, lt.k_beta))
        gamma_samples = np.zeros((sample_size, lt.k_gamma))

        beta_t = lt.beta.copy()
        gamma_t = lt.gamma.copy()

        lt_copy = deepcopy(lt)
        lt_copy.uprior[:, lt.k_beta :] = np.vstack((gamma_t, gamma_t))

        for i in range(sample_size):
            lt_copy.simulateData(beta_t, gamma_t, sim_prior=sim_prior, sim_re=sim_re)
            lt_copy.optimize(
                x0=np.hstack((beta_t, gamma_t)),
                print_level=print_level,
                max_iter=max_iter,
            )

            u_samples = lt_copy.estimateRE()

            beta_samples[i] = lt_copy.beta.copy()
            gamma_samples[i] = np.maximum(
                lt.uprior[0, lt.k_beta :],
                np.minimum(lt.uprior[1, lt.k_beta :], np.var(u_samples, axis=0)),
            )

            print(
                "sampling solution progress %0.2f" % ((i + 1) / sample_size), end="\r"
            )

        return beta_samples, gamma_samples



import numpy as np
import xarray


def saveDraws(
    beta_samples,
    u_samples,
    global_effects_names,
    random_effects_dims,
    random_effects_names,
    coords_dict={},
):
    all_coords = {}
    assert beta_samples.shape[0] == len(global_effects_names)
    beta_xr = xarray.DataArray(
        beta_samples,
        dims=("cov", "draw"),
        coords={
            "cov": global_effects_names,
            "draw": np.arange(1, beta_samples.shape[1] + 1),
        },
    )
    all_coords.update(
        {"cov": global_effects_names, "draw": np.arange(1, beta_samples.shape[1] + 1)}
    )
    assert len(u_samples) == len(random_effects_dims)
    u_xrs = []
    for i in range(len(u_samples)):
        dim_names = random_effects_dims[i] + ["draw"]
        dim_sizes = u_samples[i].shape
        assert len(dim_names) == len(dim_sizes)
        coords = {
            dim_names[j]: np.arange(1, dim_sizes[j] + 1) for j in range(len(dim_names))
        }
        coords.update(coords_dict)
        all_coords.update(coords)
        u_xrs.append(
            xarray.DataArray(u_samples[i], dims=tuple(dim_names), coords=coords)
        )
    data_vars = {
        random_effects_names[i]: u_xrs[i] for i in range(len(random_effects_names))
    }
    data_vars["beta_global"] = beta_xr
    # all_coords.update(coords_dict)

    return xarray.Dataset(data_vars, all_coords)



import numpy as np


def repeat(values, dims, full_dims):
    """
    to compute function value
    """
    assert len(values) == np.prod(dims)
    cumprod = 1
    for i in range(len(full_dims) - 1, -1, -1):
        if dims[i] == 1:
            values = np.tile(values.reshape((-1, cumprod)), (1, full_dims[i])).reshape(
                -1
            )
        cumprod *= full_dims[i]
    assert values.shape[0] == np.prod(full_dims)
    return values


def repeatTranspose(y, dims, full_dims):
    """
    to compute jacobian
    """
    values = [y]
    for i in range(len(full_dims)):
        if dims[i] == 1:
            values = [np.sum(x.reshape((full_dims[i], -1)), axis=0) for x in values]
        else:
            temp = []
            ext = temp.extend
            for x in values:
                ext(np.split(x, full_dims[i]))
            values = temp
    if len(values) == 1:
        assert np.prod(dims) == 1
        return np.array(list(values))
    assert len(np.squeeze(values)) == np.prod(dims)
    return np.squeeze(values)


def kronecker(dims, full_dims, start=0):
    """
    build Z matrix using kronecker product
    """
    # def recurse(i):
    #     if i == self.nd:
    #         return [1]
    #     if dims[i] == 1:
    #         return np.tile(recurse(i+1),(self.dimensions[i],1))
    #     else:
    #         return np.kron(np.identity(self.dimensions[i]), recurse(i+1))
    # Z = recurse(0)
    Z = [1]
    for i in range(len(dims) - 1, -1, -1):
        if dims[i] == 1:
            Z = np.tile(Z, (full_dims[i + start], 1))
        else:
            Z = np.kron(np.identity(full_dims[i + start]), Z)
    assert Z.shape[0] == np.prod(full_dims[start:])
    assert Z.shape[1] == np.prod(dims)

    return Z



import copy
from collections import namedtuple

import flme.rutils as rutils
import numpy as np
from flme.limetr import LimeTr
from spmat import BDLMat


class LME:
    def __init__(
        self,
        dimensions,
        n_grouping_dims,
        y,
        covariates,
        indicators,
        global_effects_names,
        global_intercept,
        random_effects,
    ):
        """
        Create a linear mixed effects model (LME) object
        .. math::
            {\bf y} = \beta_0 + X {\bf \beta_c} + K{\bf \beta_I} + Z \{bf u} + \epsilon
        where
            :math:`\beta_0` is a global intercept.
            :math:`X` is a matrix for global effects covariates.
            :math:`\beta_c` is a vector of covariate coefficients.
            :math:`K` is a matrix representing indicator functions
                (more details later).
            :math:`\beta_I` is a vector of indicator coefficients.
            :math:`Z` is a matrix of random effects terms (including
                      random effects covariates and random intercepts).
            :math:`u` is a vector of random effects coefficients. :math:`u` is
                      assumed to follow :math:`N(0, \Tau)` where :math:`\Tau` is
                      a diagonal matrix.
            :math:`\epsilon` is a vector of measurement noise terms.
                :math:`\epsilon` is assumed to follow :math:`N(0, \Sigma)`
                where :math:`\Sigma` is a diagonal matrix.
        To summarize, the fixed effects variables are
        .. math::
            `\beta_0, \beta_c, \beta_I, \text{diag}(\Tau), \text{diag}(\Sigma)`
        and the random effects variable is :math:`u`.
        This implementation of LME is designed for input data, including
        measurements data, covariates data, etc., that are organized by a
        certain set of dimensions, e.g. location-age-sex-year. The aforementioned
        indicator functions are indicator over sets defined by a single dimension
        or cartesian product of dimensions. For instance, the sets can be defined
        by locations, i.e. data entries from the same location belong to the one
        set, and the number of sets will be equal to the number of locations.
        The length of :math:`\beta_I` will be equal to the number of sets.
        The sets can also be defined by location-age, i.e. data entries from
        the same location and the same age group belong to one set. In this case
        the number of sets is equal to the number of locations times the number
        of age groups.
        Because of the special structure of input data, the matrices
        :math:`X, K` and their Jacobians are not formed explicitly in the
        implementation, in order to reduce memory usage.
        :math:`Z` matrix is also not formed in its full dimensions.
        Instead we only pass in its diagonal blocks, since it should have a
        block diagonal structure in general.
        The underlying core optimization engine for this program is `LimeTr`,
        which can handle a more general class of linear mixed effects models with
        built-in robust outlier detection methods.
        Args:
            dimensions (list[int]):
                a list specifying dimensions of data. The value of each component
                specifies the size of that dimension, e.g. if the data has
                dimension location-age-sex-year, the list would
                be [n_loc, n_age, n_sex, n_year].
            n_grouping_dims (int):
                an int specifying the number of dimensions on which grouping occurs,
                e.g. if ``dimensions`` is [n_loc,n_age,n_sex,n_year] and
                `n_grouping_dims` is 2, then the data will be grouped by location
                and age, and the :math:`Z` matrix corresponding to random effects
                will be a block diagonal matrix with ``n_loc`` times ``n_age``
                number of blocks.
            y (numpy.ndarray):
                One-dimensional array storing ``y`` (dependent variable) values,
                and the values must be ordered according to the ``dimensions``,
                e.g. if ``dimensions`` is [n_loc,n_age,n_sex,n_year],
                then ``y`` must be ordered by location, age, sex, year.
            covariates (dict of str: (numpy.ndarray, list[boolean])):
                A mapping that stores info of covariates. Key is the name of
                the covariate, and value is a tuple where the first component is
                a one-dimensional array that stores values of that covariate, and
                the second component specifies the dimensions of covriate
                through a boolean list, e.g. for HAQ, the tuple would be
                (values_of_haq, [True, False, False, True])
                since HAQ does not depend on age and sex.
                The dictionary would look like
                {'haq': (values_of_haq, [True, False, False, True)}
            indicators(dict of str: list[boolean]):
                A mapping for indicators.
                Key is the name of indicator, value is a boolean list that specifies
                dimensions on which to use indicators, e.g. if age-sex indicator,
                dictionary would be
                {'ind_age-sex':[False, True, True, False]}, and
                if there is an age indicator and a sex indicator, it would be
                {'ind_age':[False, True, False, False],
                 'ind_sex':[False, False, True, False]}
            global_effects_names (dict of str: list[float]):
                A mapping that stores covariates names that will be used as global effects
                and corresponding bounds on covariate multiplier values.
            global_intercept (boolean):
                A boolean indicating whether to use an intercept in global effects.
            random_effects (dict of str: (list[boolean], float)):
                A mapping that stores info of random effects. Key is the name of
                covariate or of intercept, and value is a tuple whose first element
                is a boolean list specifying dimensions on which to impose random effects,
                and the second is a value for standard deviation of the gaussian prior
                on random effects variance. When no prior it can be set to None.
                e.g. if we want to know random slope of covariate 'haq' per location-age,
                with a gaussian prior with sd of 0.1, then
                { 'haq': ([True, True, False, False], 0.1)},
                if we also want to add a random intercept on location, with no gaussian prior,
                then
                { 'haq': ([True, True, False, False], 0.1),
                  'intercept_loc': ([True, False, False, False], None)}.
                Any name that does not appear in ``covariates`` will be interpreted
                as name for an intercept. Note that the first ``n_grouping_dims``
                of the boolean list must always be True for all random effects.
        Attributes:
            beta_soln (numpy.ndarray): solution for global effects coefficients
            u_soln (list(numpy.ndarray)): solution for random effects coefficients
            yfit (numpy.ndarray): fitted y (dependent variable) values
            solve_status_msg (str): convergence info
        Raises:
            ValueError:
                If any dimension in ``dimensions`` is less than or equal to 1.
            ValueError:
                If `n_grouping_dims` is negative.
            RuntimeError:
                If value or boolean list for any covariate pair in ``covariates`` is missing.
            RuntimeError:
                If ``global_intercept`` is True and ``indicators`` is nonempty.
            RuntimeError:
                If the first ``n_grouping_dims`` of a random effect is not all True.
        """
        # if any([d <= 1 for d in dimensions]):
        #    err_msg = 'Dimensions should all be > 1.'
        #    raise ValueError(err_msg)
        if n_grouping_dims < 0:
            err_msg = "n_grouping_dims should be non-negative."
            raise ValueError(err_msg)
        self.dimensions = dimensions
        self.n_grouping_dims = n_grouping_dims
        self.n_groups = int(np.prod(dimensions[:n_grouping_dims]))
        self.grouping = [np.prod(dimensions[n_grouping_dims:])] * self.n_groups
        self.Y = y
        self.N = self.Y.shape[0]
        self.nd = len(dimensions)
        assert self.N == np.prod(self.dimensions)
        assert self.N == sum(self.grouping)

        # def bool_to_size(b):
        #     assert len(b) == self.nd
        #     for i in range(self.nd):
        #         b[i] = max(1, int(b[i])*self.dimensions[i])

        def bool_to_size(b):
            assert len(b) == self.nd
            dims = [1 for _ in range(self.nd)]
            for i in range(self.nd):
                dims[i] = max(1, int(b[i]) * self.dimensions[i])
            return dims

        self.covariates = []
        self.cov_name_to_id = {}
        i = 0
        for name, pair in covariates.items():
            if len(pair) != 2:
                err_msg = "input for " + name + "is not in correct form."
                raise RuntimeError(err_msg)
            dims = bool_to_size(pair[1])
            assert len(pair[0]) == np.prod(dims)
            self.covariates.append((pair[0], dims))
            self.cov_name_to_id[name] = i
            i += 1

        if global_intercept and indicators != {}:
            err_msg = "cannot have both global intercept be True and indicators \
                       be non-empty."
            raise RuntimeError(err_msg)
        self.global_ids = [
            self.cov_name_to_id[name] for name, bounds in global_effects_names.items()
        ]
        if len(self.global_ids) > 0:
            self.global_cov_bounds = np.transpose(
                np.array([bounds for name, bounds in global_effects_names.items()])
            )
            assert self.global_cov_bounds.shape[0] == 2
            assert all(self.global_cov_bounds[1, :] - self.global_cov_bounds[0, :] >= 0)
        else:
            self.global_cov_bounds = None
        self.global_intercept = global_intercept
        self.k_beta = len(self.global_ids) + int(self.global_intercept)
        self.indicators = []
        self.indicator_name_to_id = {}
        i = 0
        for name, ind in indicators.items():
            dims = bool_to_size(ind)
            self.k_beta += np.prod(dims)
            self.indicators.append(dims)
            self.indicator_name_to_id[name] = i
            i += 1

        self.beta_names = list(global_effects_names.keys()) + list(
            self.indicator_name_to_id.keys()
        )
        if self.global_intercept:
            self.beta_names = ["global_intercept"] + self.beta_names

        self.ran_list = []
        self.ran_eff_gamma_sd = []
        self.use_gprior = False
        for name, ran_eff_info in random_effects.items():
            ran_eff = ran_eff_info[0]
            sd = ran_eff_info[1]
            if not all(ran_eff[: self.n_grouping_dims]):
                err_msg = (
                    name
                    + ": the first "
                    + str(self.n_grouping_dims)
                    + " must be \
                           True for random effects."
                )
                raise RuntimeError(err_msg)
            dims = bool_to_size(ran_eff)
            if sd is not None:
                self.use_gprior = True
                self.ran_eff_gamma_sd.extend(
                    [sd] * np.prod(dims[self.n_grouping_dims :])
                )
            else:
                self.ran_eff_gamma_sd.extend(
                    [np.inf] * np.prod(dims[self.n_grouping_dims :])
                )
            if name in self.cov_name_to_id:
                self.ran_list.append((self.cov_name_to_id[name], dims))
            else:
                self.ran_list.append((None, dims))

        self.u_names = list(random_effects.keys())

        self.Xm = None

        return

    def X(self, beta):
        if self.Xm is not None:
            return self.Xm.dot(beta)

        assert len(beta) == self.k_beta
        y = np.zeros(self.N)
        start = 0
        if self.global_intercept:
            y += beta[start]
            start += 1
        for i in range(len(self.global_ids)):
            ind = self.global_ids[i]
            values, dims = self.covariates[ind]
            assert values.shape[0] == np.prod(dims)
            y += rutils.repeat(values, dims, self.dimensions) * beta[start + i]
        start += len(self.global_ids)
        for indicator in self.indicators:
            y += rutils.repeat(
                beta[start : start + np.prod(indicator)], indicator, self.dimensions
            )
            start += np.prod(indicator)
        return y

    def JX(self, beta):
        if self.Xm is None:
            self._buildX()
        return self.Xm

    def XT(self, y):
        if self.Xm is not None:
            return np.transpose(self.Xm).dot(y)

        assert len(y) == self.N
        start = 0
        val = []
        if self.global_intercept:
            val.append(np.sum(y))
        for i in self.global_ids:
            values, dims = self.covariates[i]
            val.append(np.dot(values, rutils.repeatTranspose(y, dims, self.dimensions)))
        for indicator in self.indicators:
            val.extend(rutils.repeatTranspose(y, indicator, self.dimensions))
        assert len(val) == self.k_beta
        return np.array(val)

    def buildZ(self, normalize=False):
        Z = []
        self.k_gamma = 0
        for ran in self.ran_list:
            id, dims = ran
            values = []
            if id is None:
                values = np.ones(self.N)
            else:
                values = rutils.repeat(
                    self.covariates[id][0], self.covariates[id][1], self.dimensions
                )
            self.k_gamma += np.prod(dims[self.n_grouping_dims :])
            Z.append(
                values.reshape((-1, 1))
                * np.tile(
                    rutils.kronecker(
                        dims[self.n_grouping_dims :],
                        self.dimensions,
                        self.n_grouping_dims,
                    ),
                    (self.n_groups, 1),
                )
            )
        if self.k_gamma > 0:
            self.Z = np.hstack(Z)
            col_norm = np.linalg.norm(self.Z, axis=0)
            if normalize:
                print("normalizing Z ...")
                print(col_norm)
                self.Z = self.Z / col_norm
            return col_norm
        else:
            self.Z = np.zeros((self.N, 1))
            return 0.0

    def _buildX(self):
        X = np.zeros((self.N, self.k_beta))
        start = 0

        if self.global_intercept is True:
            X[:, start] = np.ones(self.N)
            start += 1

        for i in range(len(self.global_ids)):
            ind = self.global_ids[i]
            values, dims = self.covariates[ind]
            assert values.shape[0] == np.prod(dims)
            X[:, start] = rutils.repeat(values, dims, self.dimensions)
            start += 1

        for indicator in self.indicators:
            X[:, start : start + np.prod(indicator)] = rutils.kronecker(
                indicator, self.dimensions, 0
            )
            start += np.prod(indicator)

        self.Xm = X

    def _solveBeta(self, S):
        if self.Xm is None:
            self._buildX()
        A = np.dot(np.transpose(self.Xm) / S**2, self.Xm)
        b = np.dot(np.transpose(self.Xm) / S, self.Y)
        return np.linalg.solve(A, b)

    def optimize(
        self,
        var=None,
        S=None,
        trim_percentage=0.0,
        share_obs_std=True,
        fit_fixed=True,
        inner_print_level=5,
        inner_max_iter=100,
        inner_tol=1e-5,
        inner_verbose=True,
        inner_acceptable_tol=1e-4,
        inner_nlp_scaling_min_value=1e-8,
        outer_verbose=False,
        outer_max_iter=1,
        outer_step_size=1,
        outer_tol=1e-6,
        normalize_Z=False,
        build_X=True,
        random_seed=0,
    ):
        """
        Run optimization routine via LimeTr.
        Args:
            var (numpy.ndarray | None, optional):
                One-dimensional array that gives initialization for variables.
                If None, the program will first run without random effects
                to obtain a starting point.
            S (numpy.ndarray | None, optional):
                One-dimensional numpy array that gives standard deviation for
                each measurement. The size of S should be the same as that of
                measurements vector. If None standard deviation of measurement
                will be treated as variables and optimized.
            trim_percentage (float | 0.0, optional):
                A float that gives percentage of datapoints to trim.
                Default is 0, i.e. no trimming.
            share_obs_std (boolean | True, optional):
                A boolean that indicates whether the model should assume data
                across studies share the same measurement standard deviation.
            fit_fixed (boolean | True, optional):
                A boolean that indicates whether to run a fit without random
                effects first in order to obtain a good starting point.
            inner_print_level (int | 5, optional):
                ``print_level`` for Ipopt.
            inner_max_iter (int | 100, optional):
                Maximum number of iterations for inner optimization.
            inner_tol (float | 1e-5, optional):
                Tolerance level for inner optimization.
            inner_verbose (boolean | True, optional):
                Verbose option for inner optimization.
            inner_acceptable_tol (float | 1e-4, optional):
                Acceptable tolerance level for inner optimization.
            inner_nlp_scaling_min_value (float | 1e-8, optional):
                Min scaling for objective function.
            outer_verbose (boolean | False, optional):
                Verbose option for outer optimization.
            outer_max_iter (int | 1, optional):
                Maximum number of iterations for outer optimization. When there
                is no trimming, outer optimization is not needed, so the default
                is set to be 1.
            outer_step_size (float |1.0, optional):
                Step size for outer optimization. Used in trimming.
            outer_tol (float | 1e-6, optional):
                Tolerance level for outer optimization.
            normalize_Z (bool | False, optional):
                Whether to normalize Z matrix before optimization.
            build_X (bool | True, optional):
                Whether to explicitly build and store X matrix.
            random_seed (int | 0, optional):
                random seed for choosing an initial point for optimization. If equals 0
                the initial point is chosen to be a vector of 0.01.
        """
        self.S = S
        self.share_obs_std = share_obs_std
        Z_norm = self.buildZ(normalize_Z)
        k = self.k_beta + self.k_gamma
        if S is None:
            if share_obs_std:
                k += 1
            else:
                k += len(self.grouping)
        print("n_groups", self.n_groups)
        print("k_beta", self.k_beta)
        print("k_gamma", self.k_gamma)
        print("total number of fixed effects variables", k)

        if self.k_gamma == 0:
            self.add_re = False
            self.k_gamma = 1
            k += 1
        else:
            self.add_re = True

        C = []
        start = self.k_beta
        for ran in self.ran_list:
            _, dims = ran
            m = np.prod(dims[self.n_grouping_dims :])
            c = np.zeros((m - 1, k))
            for i in range(m - 1):
                c[i, start + i] = 1
                c[i, start + i + 1] = -1
            C.append(c)
            start += m
        if len(C) > 0:
            self.constraints = np.vstack(C)
            assert self.constraints.shape[1] == k
        else:
            self.constraints = []

        C = None
        if len(self.constraints) > 0:

            def C(var):
                return self.constraints.dot(var)

        JC = None
        if len(self.constraints) > 0:

            def JC(var):
                return self.constraints

        c = None
        if len(self.constraints) > 0:
            c = np.zeros((2, self.constraints.shape[0]))

        self.uprior = np.array(
            [
                [-np.inf] * self.k_beta
                + [1e-7] * self.k_gamma
                + [1e-7] * (k - self.k_beta - self.k_gamma),
                [np.inf] * k,
            ]
        )

        if self.global_cov_bounds is not None:
            if self.global_intercept:
                self.uprior[:, 1 : len(self.global_ids) + 1] = self.global_cov_bounds
            else:
                self.uprior[:, : len(self.global_ids)] = self.global_cov_bounds

        self.gprior = None
        if self.use_gprior:
            assert len(self.ran_eff_gamma_sd) == self.k_gamma
            self.gprior = np.array(
                [
                    [0] * k,
                    [np.inf] * self.k_beta
                    + self.ran_eff_gamma_sd
                    + [np.inf] * (k - self.k_beta - self.k_gamma),
                ]
            )

        x0 = np.ones(k) * 0.01
        if random_seed != 0:
            np.random.seed(random_seed)
            x0 = np.random.randn(k) * 0.01
        if var is not None:
            if self.add_re is True:
                assert len(var) == k
                x0 = var
            else:
                assert len(var) == self.k_beta
                x0 = np.append(var, [1e-8])
                assert len(x0) == k

        if build_X:
            self._buildX()
        if fit_fixed or self.add_re is False:
            uprior_fixed = copy.deepcopy(self.uprior)
            uprior_fixed[:, self.k_beta : self.k_beta + self.k_gamma] = 1e-8
            if S is None or trim_percentage >= 0.01:
                model_fixed = LimeTr(
                    self.grouping,
                    int(self.k_beta),
                    int(self.k_gamma),
                    self.Y,
                    self.X,
                    self.JX,
                    self.Z,
                    S=S,
                    C=C,
                    JC=JC,
                    c=c,
                    inlier_percentage=1.0 - trim_percentage,
                    share_obs_std=share_obs_std,
                    uprior=uprior_fixed,
                )
                model_fixed.optimize(
                    x0=x0,
                    print_level=inner_print_level,
                    max_iter=inner_max_iter,
                    tol=inner_tol,
                    acceptable_tol=inner_acceptable_tol,
                    nlp_scaling_min_value=inner_nlp_scaling_min_value,
                )

                x0 = model_fixed.soln
                self.beta_fixed = model_fixed.beta
                if self.add_re is False:
                    self.beta_soln = self.beta_fixed
                    self.delta_soln = model_fixed.delta
                    self.gamma_soln = model_fixed.gamma
                    self.w_soln = model_fixed.w
                    self.info = model_fixed.info["status_msg"]
                    self.yfit_no_random = model_fixed.F(model_fixed.beta)
                    return
            else:
                self.beta_fixed = self._solveBeta(S)
                x0 = np.append(self.beta_fixed, [1e-8] * self.k_gamma)
                if self.add_re is False:
                    self.beta_soln = self.beta_fixed
                    self.yfit_no_random = self.Xm.dot(self.beta_fixed)
                    return

        model = LimeTr(
            self.grouping,
            int(self.k_beta),
            int(self.k_gamma),
            self.Y,
            self.X,
            self.JX,
            self.Z,
            S=S,
            C=C,
            JC=JC,
            c=c,
            inlier_percentage=1 - trim_percentage,
            share_obs_std=share_obs_std,
            uprior=self.uprior,
            gprior=self.gprior,
        )
        model.fitModel(
            x0=x0,
            inner_print_level=inner_print_level,
            inner_max_iter=inner_max_iter,
            inner_acceptable_tol=inner_acceptable_tol,
            inner_nlp_scaling_min_value=inner_nlp_scaling_min_value,
            inner_tol=inner_tol,
            outer_verbose=outer_verbose,
            outer_max_iter=outer_max_iter,
            outer_step_size=outer_step_size,
            outer_tol=outer_tol,
        )
        self.beta_soln = model.beta
        self.gamma_soln = model.gamma
        if normalize_Z:
            self.gamma_soln /= Z_norm**2
        self.delta_soln = model.delta
        self.info = model.info
        self.w_soln = model.w
        self.u_soln = model.estimateRE()
        self.solve_status = model.info["status"]
        self.solve_status_msg = model.info["status_msg"]

        self.yfit_no_random = model.F(model.beta)

        self.yfit = []
        Z_split = np.split(self.Z, self.n_groups)
        yfit_no_random_split = np.split(self.yfit_no_random, self.n_groups)

        for i in range(self.n_groups):
            self.yfit.append(yfit_no_random_split[i] + Z_split[i].dot(self.u_soln[i]))
        self.yfit = np.concatenate(self.yfit)
        self.model = model

        if inner_verbose == True and self.solve_status != 0:
            print(self.solve_status_msg)

    def postVarRandom(self):
        """
        y_k = X_k beta + Z_k u_k + epsilon, u has var_mat D, epsilon has var_mat R
        Var(u_k) = inv(inv(D) + Z_k'inv(R)Z_k)
        """
        assert len(self.ran_list) > 0
        Z_split = np.split(self.Z, self.n_groups)
        self.var_u = []
        S2 = []
        if self.S is None:
            if self.share_obs_std is True:
                S2 = np.ones(self.N) * self.delta_soln
            else:
                S2 = np.repeat(self.delta_soln, self.grouping)
        else:
            S2 = self.S**2
        S2_split = np.split(S2, self.n_groups)
        for i in range(self.n_groups):
            self.var_u.append(
                np.linalg.inv(
                    np.diag(1.0 / self.gamma_soln)
                    + (np.transpose(Z_split[i]) / S2_split[i]).dot(Z_split[i])
                )
            )

    def _postVarGlobal(self):
        """
        y_k = X_k beta + Z_k u_k + epsilon, u has var_mat D, epsilon has var_mat R
        Var(beta) = inv( sum_k X_k' inv(Z_k*D*Z_k' + R) X_k )
        """
        assert self.k_beta > 0
        Z_split = np.split(self.Z, self.n_groups)
        Z_split = np.split(np.zeros((self.N, self.k_gamma)), self.n_groups)

        self.var_beta = np.zeros((self.k_beta, self.k_beta))

        X = np.zeros((self.N, self.k_beta))
        start = 0

        if self.global_intercept is True:
            X[:, start] = np.ones(self.N)
            start += 1

        for i in range(len(self.global_ids)):
            ind = self.global_ids[i]
            values, dims = self.covariates[ind]
            assert values.shape[0] == np.prod(dims)
            X[:, start] = rutils.repeat(values, dims, self.dimensions)
            start += 1

        for indicator in self.indicators:
            X[:, start : start + np.prod(indicator)] = rutils.kronecker(
                indicator, self.dimensions, 0
            )
            start += np.prod(indicator)

        X_split = np.split(X, self.n_groups)

        S2 = []
        if self.S is None:
            if self.share_obs_std is True:
                S2 = np.ones(self.N) * self.delta_soln
            else:
                S2 = np.repeat(self.delta_soln, self.grouping)
        else:
            S2 = self.S**2
        S2_split = np.split(S2, self.n_groups)

        for i in range(self.n_groups):
            V = Z_split[i].dot(np.diag(self.gamma_soln)).dot(
                np.transpose(Z_split[i])
            ) + S2_split[i] * np.identity(self.grouping[i])
            self.var_beta += (
                np.transpose(X_split[i]).dot(np.linalg.inv(V)).dot(X_split[i])
            )
        self.var_beta = np.linalg.inv(self.var_beta)

    def postVarGlobal(self):
        assert self.k_beta > 0

        # X = np.zeros((self.N, self.k_beta))
        # start = 0
        #
        # if self.global_intercept is True:
        #     X[:,start] = np.ones(self.N)
        #     start += 1
        #
        # for i in range(len(self.global_ids)):
        #     ind = self.global_ids[i]
        #     values, dims = self.covariates[ind]
        #     assert values.shape[0] == np.prod(dims)
        #     X[:,start] = rutils.repeat(values, dims, self.dimensions)
        #     start += 1
        #
        # for indicator in self.indicators:
        #     X[:, start:start + np.prod(indicator)] = rutils.kronecker(indicator, self.dimensions, 0)
        #     start += np.prod(indicator)

        if self.Xm is None:
            self._buildX()

        X = self.Xm

        S2 = []
        if self.S is None:
            if self.share_obs_std is True:
                S2 = np.ones(self.N) * self.delta_soln
            else:
                S2 = np.repeat(self.delta_soln, self.grouping)
        else:
            S2 = self.S**2

        mat = BDLMat(
            S2,
            np.zeros((self.N, self.k_gamma)) * np.sqrt(self.gamma_soln),
            self.grouping,
        )
        self.var_beta = np.dot(np.transpose(X), mat.invdot(X))
        self.var_beta = np.linalg.inv(self.var_beta)

    def sampleGlobalWithLimeTr(self, sample_size=100, max_iter=300):
        beta_samples, gamma_samples = LimeTr.sampleSoln(
            self.model, sample_size=sample_size, max_iter=max_iter
        )
        return beta_samples, gamma_samples

    def _drawBeta(self, n_draws):
        if self.global_cov_bounds is None:
            return np.transpose(
                np.random.multivariate_normal(self.beta_soln, self.var_beta, n_draws)
            )

        bounds = self.uprior[:, : self.k_beta]
        # print(bounds)
        beta_samples = np.empty((self.k_beta, 0), float)
        while beta_samples.shape[1] < n_draws:
            samples = np.transpose(
                np.random.multivariate_normal(self.beta_soln, self.var_beta, n_draws)
            )
            beta_samples = np.hstack(
                (
                    beta_samples,
                    samples[
                        :,
                        np.all(
                            (samples - bounds[0, :].reshape((-1, 1)) >= 0)
                            & (samples - bounds[1, :].reshape((-1, 1)) <= 0),
                            axis=0,
                        ),
                    ],
                )
            )
        return beta_samples[:, :n_draws]

    def draw(self, n_draws=10):
        """
        Draw samples of global effects coefficient beta and random intercepts u
        from their corresponding posterior distribution.
        Args:
            n_draws(int | 10): number of draws
        Returns:
            list(numpy.ndarray):
                Arrays for beta samples and u samples
        """
        beta_samples = []
        if self.k_beta > 0:
            beta_samples = self._drawBeta(n_draws)
        u_samples = [[] for _ in range(len(self.ran_list))]
        if self.add_re is True:
            for i in range(self.n_groups):
                # sample all random effects u in global group i
                samples = np.random.multivariate_normal(
                    self.u_soln[i], self.var_u[i], n_draws
                )
                start = 0
                for j in range(len(self.ran_list)):
                    ran_eff = self.ran_list[j]
                    dims = ran_eff[1]
                    # extract u related to random effect j
                    u_samples[j].append(
                        samples[
                            :, start : start + np.prod(dims[self.n_grouping_dims :])
                        ].reshape((n_draws, -1))
                    )
                    start += np.prod(dims[self.n_grouping_dims :])
            for i in range(len(u_samples)):
                # each u_sample is a matrix of dimension n_draws-by-n_groups specific
                # to that random effect (>= number of global groups)
                # and the matrix is then transposed
                u_samples[i] = np.transpose(np.hstack(u_samples[i]))
            for i in range(len(self.ran_list)):
                ran_eff = self.ran_list[i]
                _, dims = ran_eff
                u_samples[i] = u_samples[i].reshape(tuple(dims + [n_draws])).squeeze()
        # each u_samples[i] have different shapes
        return beta_samples, u_samples

    def outputDraws(self, n_draws=10, by_type=True, combine_cov=True):
        """
        Output draws as namedtuples. Each tuple has two fields, `array` and `name(s)`.
        `array` stores a numpy array of draws. There are three types of estimated
        values we need draws: global covariates, indicators, and random effects.
        For all three types, the dimension of `array` is equal to the original
        dimension of the estimate, plus one extra dimension whose size is the number of draws.
        e.g.
        for covariates, its original dimension is 1, since it is a scalar,
        thus the `array` dimension is n_draws;
        for indicator, age-sex as an example. its original dimension is n_age-by-n_sex,
        thus the `array` dimension is n_age-by-n_sex-by-n_draws;
        random_effects is similar to indicators.
        Args:
            n_draws (int | 10): number of draws
            by_type (boolean | True):
                whether to group samples according to the three types
            combine_cov (boolean | True):
                whether to combine covariates into one namedtuple
        Returns:
            list(namedtuple):
                List of namedtuples. Each tuple is length two. The first component
                is a numpy array of draws, and the second is the name of the estimate.
                If ``by_type`` is true, estimates are grouped by type.
                If ``combine_cov`` is true, global covariates are combined into one
                namedtuple.
        """
        beta_samples, u_samples = self.draw(n_draws)
        samples = []
        n_cov = int(self.global_intercept) + len(self.global_ids)
        for i in range(n_cov):
            samples.append(beta_samples[i, :])
        start = n_cov
        for dim in self.indicators:
            samples.append(
                beta_samples[start : start + np.prod(dim), :]
                .reshape(tuple(dim + [n_draws]))
                .squeeze()
            )
            start += np.prod(dim)
        assert start == self.k_beta
        assert len(samples) == len(self.beta_names)
        assert len(u_samples) == len(self.u_names)

        samples_name_pairs = zip(samples + u_samples, self.beta_names + self.u_names)
        Draws = namedtuple("Draws", "array, name")
        samples_name_pairs = [Draws(*pair) for pair in samples_name_pairs]

        if by_type is True:
            cov_samples = samples_name_pairs[:n_cov]
            indicator_samples = samples_name_pairs[n_cov : len(self.beta_names)]
            raneff_samples = samples_name_pairs[len(self.beta_names) :]
            if combine_cov is True:
                if n_cov > 0:
                    CovDraws = namedtuple("CovDraws", "array, names")
                    cov_samples = CovDraws(
                        beta_samples[:n_cov, :], self.beta_names[:n_cov]
                    )
                else:
                    cov_samples = None
            return cov_samples, indicator_samples, raneff_samples

        return samples_name_pairs
